#!/bin/bash

# Download mercartor-tile-based maps
# Last change: 2008-05-09 (Frank Baettermann)

# Tools:
# - bc
# - curl
# - file
# - montage (image-magick)

if [ $# -eq 0 ]; then
    echo "Usage: $0 [OPTIONS]"
    echo "  --target:map"
    echo "          openstreetmap (default)[:tah (default)]"
    echo "  --tiles X1,Y1,X2,Y2,ZOOM   tile-area (range depends on zoom!)"
    echo "  --montage                   combine tiles to single image"
    echo "  --output-dir DIR            output directory (should not exist)"
    echo "  --proxy [USER[:PASSWORD]@]HOST[:PORT]"
    echo "  --proxy-type TYPE           http (default) | socks4 | socks5"
    exit 1
fi

# defaults
TARGET="openstreetmap"
TARGET_MAP=""
CURL_PROXY=""
CURL_PROXY_TYPE="http"
CURL_PROXY_USER=""
CURL_USER_AGENT="Mozilla/5.0"
MAP_X1=-1
MAP_Y1=-1
MAP_X2=-1
MAP_Y2=-1
MAP_ZOOM=0
MONTAGE="NO"
MONTAGE_FILENAME="montage"
MONTAGE_PARAMS=""
OUTPUT_DIR=""

# parse parameters
while [ $# -gt 0 ]; do
    case "$1" in
        "--target")
            if [ `echo $2 | grep ':' | wc -l` -ne 0 ]; then
                TARGET=`echo $2 | cut -d ':' -f 1`
                TARGET_MAP=`echo $2 | cut -d ':' -f 2`
            else
                TARGET=$2
            fi
            shift
            ;;
        "--tiles")
            MAP_X1=`echo $2 | cut -d "," -f 1`
            MAP_Y1=`echo $2 | cut -d "," -f 2`
            MAP_X2=`echo $2 | cut -d "," -f 3`
            MAP_Y2=`echo $2 | cut -d "," -f 4`
            MAP_ZOOM=`echo $2 | cut -d "," -f 5`
            shift
            ;;
        "--montage")
            MONTAGE="YES"
            ;;
        "--output-dir")
            OUTPUT_DIR=$2
            shift
            ;;
        "--proxy")
            if [ `echo $2 | grep '@' | wc -l` -ne 0 ]; then
                CURL_PROXY_USER=`echo $2 | cut -d '@' -f 1`
                CURL_PROXY=`echo $2 | cut -d '@' -f 2`
            else
                CURL_PROXY=$2
            fi
            shift
            ;;
        "--proxy-type")
            CURL_PROXY_TYPE=$2
            shift
            ;;
        *)
            echo "Unknown parameter: $1"
            exit 1
            ;;
    esac
    shift
done

# check target
case "$TARGET" in
    "openstreetmap")
        if [ -z "$TARGET_MAP" ]; then
            TARGET_MAP="tah"
        fi
        # check target-map
        case "$TARGET_MAP" in
# !!! seems like the mapnik-tiles are spread among the mirrors and none of them has all !!!
# (we would have to consider this when buildung the tile-url...)
#             "osm")
#                 MAP_ZOOM_MAX=18
#                 URL_SERVER=('http://a.tile.openstreetmap.org/' 'http://c.tah.openstreetmap.org/' 'http://c.tah.openstreetmap.org/')
#                 URL_TILE='{ZOOM}/{X}/{Y}.png'
#                 ;;
            "tah")
                MAP_ZOOM_MAX=17
                URL_SERVER=('http://a.tah.openstreetmap.org/' 'http://c.tah.openstreetmap.org/' 'http://c.tah.openstreetmap.org/')
                URL_TILE='Tiles/tile/{ZOOM}/{X}/{Y}.png'
                ;;
            *)
                echo "Unknown target-map: $TARGET_MAP"
                exit 1
                ;;
        esac
        TILE_SIZE=256
        TILE_FILE_EXT='png'
        MONTAGE_PARAMS="-quality 100"
        ;;
    *)
        echo "Unknown target: $TARGET"
        exit 1
        ;;
esac

# compute map-size
MAP_SIZE=$((2**$MAP_ZOOM))

# check tile coords
if [ $MAP_X1 -lt 0 ] || [ $MAP_X1 -gt $MAP_X2 ] || [ $MAP_X2 -ge $MAP_SIZE ] \
    || [ $MAP_Y1 -lt 0 ] || [ $MAP_Y1 -gt $MAP_Y2 ] || [ $MAP_Y2 -ge $MAP_SIZE ]; then
    echo "Error: Wrong tile coordinates: $MAP_X1,$MAP_Y1,$MAP_X2,$MAP_Y2"
    exit 1
fi

# build curl parameters
CURL_PARAMS=""
if [ -n "$CURL_PROXY" ]; then
    case "$CURL_PROXY_TYPE" in
        "http")
            CURL_PARAMS="$CURL_PARAMS --proxy $CURL_PROXY"
            ;;
        "socks4")
            CURL_PARAMS="$CURL_PARAMS --socks4 $CURL_PROXY"
            ;;
        "socks5")
            CURL_PARAMS="$CURL_PARAMS --socks5 $CURL_PROXY"
            ;;
        *)
            echo "Error: Unknown proxy-type: $CURL_PROXY_TYPE"
            exit 1
            ;;
    esac
    if [ -n "$CURL_PROXY_USER" ]; then
        CURL_PARAMS="$CURL_PARAMS --proxy-user $CURL_PROXY_USER"
    fi
fi
if [ -n "$CURL_USER_AGENT" ]; then
    CURL_PARAMS="$CURL_PARAMS --user-agent $CURL_USER_AGENT"
fi

# create and enter output directory
if [ -z "$OUTPUT_DIR" ]; then
    OUTPUT_DIR="`date +%Y%m%d%H%M%S`-${TARGET}-${TARGET_MAP}"
fi
if [ -d "$OUTPUT_DIR" ]; then
    echo "Output directory already exists."
    exit 1
fi
mkdir $OUTPUT_DIR &> /dev/null
if [ $? -ne 0 ]; then
    echo "Could not create output directory. Check permissions!"
    exit 1
fi
cd $OUTPUT_DIR

#function definitions

function pad_zero()
{
    local OUT="000000${1}"
    echo ${OUT:$((${#OUT}-6)):6}
}

function tile_url()
{
    # select server randomly
    local SERVER="${URL_SERVER[$(($RANDOM % ${#URL_SERVER[*]}))]}"
    local TILE="${URL_TILE}"
    TILE="${TILE/'{X}'/$X}"
    TILE="${TILE/'{Y}'/$Y}"
    TILE="${TILE/'{ZOOM}'/$MAP_ZOOM}"
    echo "${SERVER}${TILE}"
}

function tile_filename()
{
    echo "tile-`pad_zero $Y`-`pad_zero $X`.$TILE_FILE_EXT"
}

function tile_download()
{
    local RETURN=0
    local RETRIES=0
    while [ $RETRIES -lt 5 ]; do
	curl \
	    $CURL_PARAMS \
	    --max-time 10 \
	    --output `tile_filename` \
	    `tile_url` \
	    >> curl.log 2>> curl.log
	RETURN=$?
	if [ $RETURN -eq 0 ]; then
	    return 0
	else
	    # error or timeout
	    ((RETRIES++))
	fi
    done
    # error
    return $RETURN
}

# perform downloads
for Y in `seq $MAP_Y1 $MAP_Y2`; do
    for X in `seq $MAP_X1 $MAP_X2`; do
        echo -n "Tile ${X}x${Y} ... "
	tile_download
        if [ $? -ne 0 ]; then
	    echo "Error (see curl.log)"
	    exit 1
	else
    	    echo "Ok"
        fi
    done
done

# combine tiles
if [ $MONTAGE = "YES" ]; then
    echo -n "Montage... "

    # -tile wxh: tile alignment
    montage \
        -tile $(($MAP_X2 - $MAP_X1 + 1))x$(($MAP_Y2 - $MAP_Y1 + 1)) \
        -geometry ${TILE_SIZE}x${TILE_SIZE} \
        $MONTAGE_PARAMS \
        tile-*${TILE_FILE_EXT} \
        $MONTAGE_FILENAME.${TILE_FILE_EXT} \
        &> montage.log

    if [ $? -eq 0 ]; then
        echo "Ok"
    else
        echo "Error (see montage.log)"
    fi
fi
